// -*- C++ -*-
/*!
 * @file  rtc_factorization.cpp
 * @brief Image processing of the factorization
 * @date $Date$
 *
 * @author EAST JAPAN ACCOUNTING CENTER CO.,LTD.
 *
 * BSD 3-clause "New" or "Revised" License
 *
 * $Id$
 */

#include "rtc_factorization.h"
#include "Matrix_factorization.h"

// Module specification
// <rtc-template block="module_spec">
static const char* rtc_factorization_spec[] =
  {
    "implementation_id", "rtc_factorization",
    "type_name",         "rtc_factorization",
    "description",       "Image processing of the factorization",
    "version",           "1.1",
    "vendor",            "EAC",
    "category",          "Image",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
rtc_factorization::rtc_factorization(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_in_img1In("ImgIn_2dcdp1", m_in_img1),
    m_in_img2In("ImgIn_2dcdp2", m_in_img2),
    m_rlt_2dcdpIn("Result_2dcdp", m_rlt_2dcdp),
    m_trj_filenoIn("Traject_FileNo", m_trj_fileno),
    m_out_imgOut("ImgOut", m_out_img),
    m_rlt_objOut("Result_factorization", m_rlt_obj),
    m_out_trjOut("TrajectOut", m_out_trj),
    m_out_trjpathOut("TrajectPath", m_out_trjpath)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
rtc_factorization::~rtc_factorization()
{
  // add to code
  nvdRelease();
}


/*!
 * Initialization of the factorization RTC
 */
RTC::ReturnCode_t rtc_factorization::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("ImgIn_2dcdp1", m_in_img1In);
  addInPort("ImgIn_2dcdp2", m_in_img2In);
  addInPort("Result_2dcdp", m_rlt_2dcdpIn);
  addInPort("Traject_FileNo", m_trj_filenoIn);
  
  // Set OutPort buffer
  addOutPort("ImgOut", m_out_imgOut);
  addOutPort("Result_factorization", m_rlt_objOut);
  addOutPort("TrajectOut", m_out_trjOut);
  addOutPort("TrajectPath", m_out_trjpathOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
  
  // add to code
  nvdFactorOnInitialize();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_factorization::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_factorization::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_factorization::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Activate the factorization RTC
 */

RTC::ReturnCode_t rtc_factorization::onActivated(RTC::UniqueId ec_id)
{
  // add to code
  nvdFactorOnActivated();

  return RTC::RTC_OK;
}

/*!
 * Deactivate the factorization RTC
 */

RTC::ReturnCode_t rtc_factorization::onDeactivated(RTC::UniqueId ec_id)
{
  // add to code
  nvdFactorOnDeactivated();

  return RTC::RTC_OK;
}

/*!
 * Periodic task of the factorization RTC
 */

RTC::ReturnCode_t rtc_factorization::onExecute(RTC::UniqueId ec_id)
{
  // add to code
  nvdFactorOnExecute();

  return RTC::RTC_OK;
}

/*!
 * onAborting the factorization RTC
 */

RTC::ReturnCode_t rtc_factorization::onAborting(RTC::UniqueId ec_id)
{
  // add to code
  nvdFactorOnAborting();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_factorization::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Reset initialization of the factorization RTC
 */

RTC::ReturnCode_t rtc_factorization::onReset(RTC::UniqueId ec_id)
{
  // add to code start
  nvdFactorOnReset();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_factorization::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_factorization::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/***************************************************************************************************
 *  Function        : nvdFactorOnInitialize
 *  Description     : When onInitialize, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdFactorOnInitialize( void )
{
  std::cout << _ON_INITIALIZE << std::endl;

  DWORD result = 0;

  char_type strPath[DEF_MAX_LENGTH];
  char_type strDrive[DEF_MIN_LENGTH];
  char_type strDir[DEF_MAX_LENGTH];
  char_type strFile[DEF_MID_LENGTH];
  char_type strExt[DEF_MIN_LENGTH];

  memset(strPath, 0x00, sizeof(strPath));
  memset(strDrive, 0x00, sizeof(strDrive));
  memset(strDir, 0x00, sizeof(strDir));
  memset(strFile, 0x00, sizeof(strFile));
  memset(strExt, 0x00, sizeof(strExt));
  memset(g_absolute_dir, 0x00, sizeof(g_absolute_dir));
  
  result = GetModuleFileName(NULL, strPath, sizeof(strPath));
  if (result == DEF_GET_PATH_NG)
  {
    char_type esc[] = "\\";
    std::cout << _WARNING_PATH_CREATION << std::endl;
    GetFullPathName(".", DEF_MAX_LENGTH, g_absolute_dir, NULL);
    strcat(g_absolute_dir, esc);
  }
  else
  {
    _splitpath(strPath, strDrive, strDir, strFile, strExt);
    sprintf(g_absolute_dir, "%s%s", strDrive, strDir);
  }
  std::cout << "absolute dir path: " << g_absolute_dir << std::endl;
}

/***************************************************************************************************
 *  Function        : nvdFactorOnActivated
 *  Description     : When onActivated, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdFactorOnActivated( void )
{
  std::cout << _ON_ACTIVATED << std::endl;

  m_in_img1.images.length( IMAGES_LENGTH_2 );
  m_in_img2.images.length( IMAGES_LENGTH_2 );
  m_out_img.images.length( IMAGES_LENGTH_2 );

  m_out_trjpath.data.length(NUM_IMAGE);

  // variables of matrix
  m_uhwInEvent    = NO_EVENT;
  m_uhwMatrixMode = M_MODE_INIT;

  nvdSetStageControl( FACTOR_STAGE1 );
  m_ubStageCtrlBackUp = FACTOR_STAGE_MAX;

  nvdSetErrorCode( ERR_CODE_INIT );

  inp.Create( XIMG_320, YIMG_240 );
  ref.Create( XIMG_120, YIMG_90 );
}

/***************************************************************************************************
 *  Function        : nvdFactorOnDeactivated
 *  Description     : When onDeactivated, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdFactorOnDeactivated( void )
{
  std::cout << _ON_DEACTIVATED << std::endl;

  m_in_img1.images.release();
  m_in_img2.images.release();
  m_out_img.images.release();

  m_out_trjpath.data.release();

  ref.Release();
  inp.Release();

  g_obj_mesh.Reset();
}

/***************************************************************************************************
 *  Function        : nvdFactorOnAborting
 *  Description     : When onAborting, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdFactorOnAborting( void )
{
  std::cout << _ON_ABORTING << std::endl;
  std::cout << _ERROR_RESET << std::endl;
}

/***************************************************************************************************
 *  Function        : nvdFactorOnReset
 *  Description     : When onReset, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdFactorOnReset( void )
{
  std::cout << _ON_RESET << std::endl;
}

/***************************************************************************************************
 *  Function        : nvdFactorOnExecute
 *  Description     : This function execute the main process of Factorization RTC.
 *  Argument        : m_ubStageCtrl
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdFactorOnExecute( void )
{
  uword_type auwReturn = RESULT_NG;

  switch ( m_ubStageCtrl )
  {
    case FACTOR_STAGE1:
      nvdSetErrorCode( ERR_CODE_INIT );
      auwReturn = uwMainMatrixExecute();
      if( auwReturn == RESULT_OK )
      {
        nvdSetStageControl( FACTOR_STAGE2 );
      }
      break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
    case FACTOR_STAGE2:
      nvdSetErrorCode( ERR_CODE_INIT );
      if( bCreateMesh( g_obj_mesh ) )
      {
        std::cout << FACTORIZATION_SUCCESS << std::endl;
        nvdSetStageControl( FACTOR_STAGE1 );
      }
      else
      {
        nvdSetStageControl( ERROR_DETECTED );
        nvdSetErrorCode( ERR_OBJ_CREATE );
      }
      break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
    case ERROR_DETECTED:
      if( m_ubStageCtrlBackUp != ERROR_DETECTED )
      {
        nvdOutputErrorMessages();
      }

      break;
    //---------------------------------------------------------------------------------

    default:
      std::cout << _STAGE_ERROR << std::endl;
      break;
  }

  m_ubStageCtrlBackUp = m_ubStageCtrl;

}

/***************************************************************************************************
 *  Function        : uwMainMatrixExecute
 *  Description     : This function the main processing of the matrix.
 *  Argument        : event, matrix mode
 *  Return          : matrix mode
 **************************************************************************************************/
uword_type rtc_factorization::uwMainMatrixExecute( void )
{
  uword_type auwReturn = RESULT_NG;
  /* Analyze the input event of the matrix */
  uhword_type auhwInEvent = uhwInEvCheckofMatrix();

  if( (auhwInEvent < IN_EVENT_MAX) && (m_uhwMatrixMode < M_MODE_MAX) )
  {
    m_uhwInEvent = auhwInEvent;
    nvdModeMatrixExecute( m_uhwInEvent, m_uhwMatrixMode );
    if( m_uhwMatrixMode == INPUT_DATA_COMP )
    {
      m_uhwMatrixMode = M_MODE_INIT;
      nvdOutCameraImagesCreation();
      auwReturn = RESULT_OK;
    }
  }

  return auwReturn;

}

/***************************************************************************************************
 *  Function        : uhwInEvCheckofMatrix
 *  Description     : This function check the event of the matrix.
 *  Argument        : event
 *  Return          : event of the matrix
 **************************************************************************************************/
uhword_type rtc_factorization::uhwInEvCheckofMatrix( void )
{
  uhword_type auhwInEvent = uhwCheckofInputImage();

  if( auhwInEvent != EVENT_IMAGE )
  {
    auhwInEvent = uhwCheckofResult2dcdp();
  }

  return auhwInEvent;

}

/***************************************************************************************************
 *  Function        : uhwCheckofResult2dcdp
 *  Description     : This function checks the inputs 2dcdp result.
 *  Argument        : Input port of Result2dcdp
 *  Return          : evevt
 **************************************************************************************************/
uhword_type rtc_factorization::uhwCheckofResult2dcdp( void )
{
  uhword_type auhwInEvent = NO_EVENT;

  if( m_rlt_2dcdpIn.isNew() )
  {
    m_rlt_2dcdpIn.read();
    if( m_rlt_2dcdp.rtc_id == RESULT_2DCDP_1 )
    {
      auhwInEvent = EVENT_RESULT1;
    }
    else if( m_rlt_2dcdp.rtc_id == RESULT_2DCDP_2 )
    {
      auhwInEvent = EVENT_RESULT2;
    }
    else{}
  }

  return auhwInEvent;
}

/***************************************************************************************************
 *  Function        : uhwCheckofInputImage
 *  Description     : This function checks the inputs image data.
 *  Argument        : Input port of image data
 *  Return          : evevt
 **************************************************************************************************/
uhword_type rtc_factorization::uhwCheckofInputImage( void )
{
  uhword_type auhwInEvent = NO_EVENT;

  m_uhwImageData = KIND_IMAGE_INIT;
  if( m_in_img1In.isNew() )
  {
    m_in_img1In.read();
    if( bCheckMultiCameraImages(m_in_img1) )
    {
      m_uhwImageData = KIND_IMAGE_DATA1;
      auhwInEvent = EVENT_IMAGE;
    }
  }
  else
  {
    if( m_in_img2In.isNew() )
    {
      m_in_img2In.read();
      if( bCheckMultiCameraImages(m_in_img2) )
      {
        m_uhwImageData = KIND_IMAGE_DATA2;
        auhwInEvent = EVENT_IMAGE;
      }
    }
  }

  return auhwInEvent;
}

/***************************************************************************************************
 *  Function        : bCheckMultiCameraImages
 *  Description     : This function will check whether the mult-cameraimages data is valid.
 *  Argument        : multi-cameraimages
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_factorization::bCheckMultiCameraImages( RTC::MultiCameraImages MultiImage )
{
  bool_type abReturn = FALSE;

  uword_type imgwidth;
  uword_type imgheight;
  ubyte_type result[2];

  ubyte_type i = 0;
  for( i = 0; i < IMAGES_LENGTH_2; i++ )
  {
    imgwidth  = MultiImage.images[i].width;
    imgheight = MultiImage.images[i].height;

    result[i] = FALSE;
    if( bCheckMultiCameraImagesOfSize(imgwidth, imgheight) )
    {
      if( MultiImage.images[i].bpp == FORMAT_COLOR24 )
      {
        result[i] = TRUE;
      }
      else
      {
        std::cout << _ERROR_IMAGE_COLOR << std::endl;
      }
    }
  }

  if( (result[0] == TRUE) && (result[1] == TRUE) )
  {
    abReturn = TRUE;
  }
  else
  {
    nvdSetStageControl( ERROR_DETECTED );
    nvdSetErrorCode( ERR_IMAGE_DATA );
  }

  return abReturn;
}

/***************************************************************************************************
 *  Function        : bCheckMultiCameraImagesOfSize
 *  Description     : This function will check the range of the cameraimage data.
 *  Argument        : cameraimage
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type rtc_factorization::bCheckMultiCameraImagesOfSize( uword_type imgwidth, uword_type imgheight )
{
  bool_type abReturn = FALSE;

  if( (imgwidth == 0) || (imgheight == 0) )
  {
    std::cout << _ERROR_IMAGE_EMPTY << std::endl;
  }
  else if( (imgwidth <= XIMG_320) && (imgheight <= YIMG_240) )
  {
    abReturn = TRUE;
  }
  else
  {
    std::cout << _ERROR_IMAGE_SIZE << std::endl;
  }

  return abReturn;
}

/***************************************************************************************************
 *  Function        : nvdModeMatrixExecute
 *  Description     : This function matrix execution of matrix.
 *  Argument        : event, matrix mode
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdModeMatrixExecute( uhword_type auhwInEvent, uhword_type auhwMatrixMode )
{
  uhword_type auhwNewState;
  uhword_type auhwProcID;

  uhword_type(rtc_factorization::*uhwProcess_tbl[])(void) = {
    &rtc_factorization::uhwProc_InputImage   ,	/*  0	PROC_IMAG */
    &rtc_factorization::uhwProc_InputResult1 ,	/*  1	PROC_RLT1 */
    &rtc_factorization::uhwProc_InputResult2 ,	/*  2	PROC_RLT2 */
  };

  auhwNewState = nstInEvevtCheck_tbl[auhwInEvent][auhwMatrixMode].uhwNextState;
  auhwProcID   = nstInEvevtCheck_tbl[auhwInEvent][auhwMatrixMode].uhwProcessCode;

  if( auhwNewState != MT_NO_CHG )
  {
    m_uhwMatrixMode = auhwNewState;
  }

  if( auhwProcID != NO_PROC )
  {
    if( (auhwProcID >= PROC_MIN) && (auhwProcID < PROC_MAX) )
    {
      auhwNewState = (this->*uhwProcess_tbl[auhwProcID])();
      if( auhwNewState != MT_NO_CHG )
      {
        m_uhwMatrixMode = auhwNewState;
      }
    }
  }
}

/***************************************************************************************************
 *  Function        : uhwProc_InputImage
 *  Description     : This function saves the image data.
 *  Argument        : image data
 *  Return          : next matrix mode
 **************************************************************************************************/
uhword_type rtc_factorization::uhwProc_InputImage( void )
{
  uhword_type auhwRetCode = MT_NO_CHG;

  if( m_uhwImageData == KIND_IMAGE_DATA1 )
  {
    nvdSaveImageData( &m_in_img1 );
    m_uhwImageData = KIND_IMAGE_INIT;
  }
  else if( m_uhwImageData == KIND_IMAGE_DATA2 )
  {
    nvdSaveImageData( &m_in_img2 );
    m_uhwImageData = KIND_IMAGE_INIT;
  }
  else{}

  return auhwRetCode;

}

/***************************************************************************************************
 *  Function        : uhwProc_InputResult1
 *  Description     : This function saves the 2dcdp1 result data.
 *  Argument        : 2dcdp1 result data
 *  Return          : next matrix mode
 **************************************************************************************************/
uhword_type rtc_factorization::uhwProc_InputResult1( void )
{
  uhword_type auhwRetCode = MT_NO_CHG;
  bool_type abFlag = FALSE;

  uword_type inp_len = m_rlt_2dcdp.inp_len;
  uword_type ref_len = m_rlt_2dcdp.ref_len;
  uword_type inp_w   = m_rlt_2dcdp.inp_w;
  uword_type ref_w   = m_rlt_2dcdp.ref_w;

  if( bAllocateMemory_m_corsp(ref_len) )
  {
    if( bAllocateMemory_pickbuf(inp_len) )
    {
      if( bAllocateMemory_pickup1(ref_len) )
      {
        nvdCopy_pickup1(ref_len);

        inpW = inp_w;
        refW = ref_w;
        if( inp_len != 0 ) inpH = (inp_len / inp_w);
        if( ref_len != 0 ) refH = (ref_len / ref_w);

        abFlag = TRUE;
      }
    }
  }

  if( abFlag != TRUE )
  {
    nvdSetStageControl( ERROR_DETECTED );
    nvdSetErrorCode( ERR_ALLOCATE_M );
  }

  return auhwRetCode;

}

/***************************************************************************************************
 *  Function        : uhwProc_InputResult2
 *  Description     : This function saves the 2dcdp2 result data.
 *  Argument        : 2dcdp2 result data
 *  Return          : next matrix mode
 **************************************************************************************************/
uhword_type rtc_factorization::uhwProc_InputResult2( void )
{
  uhword_type auhwRetCode = MT_NO_CHG;

  uword_type ref_len = m_rlt_2dcdp.ref_len;

  if( bAllocateMemory_pickup2(ref_len) )
  {
    nvdCopy_pickup2(ref_len);
  }
  else
  {
    nvdSetStageControl( ERROR_DETECTED );
    nvdSetErrorCode( ERR_ALLOCATE_M );
  }

  return auhwRetCode;

}

/***************************************************************************************************
 *  Function        : nvdSaveImageData
 *  Description     : This function saves image data.
 *  Argument        : *image_p
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdSaveImageData( RTC::MultiCameraImages *image_p )
{
  cv::Mat m_image;

  m_image = cv::Mat( cv::Size( image_p->images[IMG_POS_C].width, image_p->images[IMG_POS_C].height ), CV_8UC3 );
  memcpy( m_image.data, (void *)&(image_p->images[IMG_POS_C].pixels[0]), image_p->images[IMG_POS_C].pixels.length() );
  m_IplImage = m_image;

  cvResize( &m_IplImage, ref.iplImage(), CDP_IMAGE_RESIZE_TYPE );

  m_image = cv::Mat( cv::Size( image_p->images[IMG_INP_C].width, image_p->images[IMG_INP_C].height ), CV_8UC3 );
  memcpy( m_image.data, (void *)&(image_p->images[IMG_INP_C].pixels[0]), image_p->images[IMG_INP_C].pixels.length() );
  m_IplImage = m_image;

  cvResize( &m_IplImage, inp.iplImage(), CDP_IMAGE_RESIZE_TYPE );

}

/***************************************************************************************************
 *  Function        : nvdOutCameraImagesCreation
 *  Description     : This function sets the data to the output port.
 *  Argument        : nothing special
 *  Return          : data of output port
 **************************************************************************************************/
void rtc_factorization::nvdOutCameraImagesCreation( void )
{
  bool_type abWrtResult = TRUE;
  cv::Mat m_image;

  m_image = cv::cvarrToMat( ref.iplImage() );
  if( RESULT_OK == uwSetCameraImages( m_image, SEQ_NO_1 ) )
  {
    m_image = cv::cvarrToMat( inp.iplImage() );
    if( RESULT_OK == uwSetCameraImages( m_image, SEQ_NO_2 ) )
    {
      abWrtResult = m_out_imgOut.write();	
      if( abWrtResult == FALSE )
      {
        std::cout << _ERROR_WRITE_IMG << std::endl;
      }
    }
  }
}

/***************************************************************************************************
 *  Function        : uwSetCameraImages
 *  Description     : This function sets the data of the Multi-CameraImage type.
 *  Argument        : Mat type dsp_image, sequence number
 *  Return          : information of normal end or abnormal end.
 **************************************************************************************************/
uword_type rtc_factorization::uwSetCameraImages( cv::Mat dsp_image, uword_type seq_num )
{
  uword_type auwReturn = RESULT_NG;

  if( dsp_image.empty() )
  {
    std::cout << _ERROR_IMAGE_EMPTY << std::endl;
  }
  else
  {
    sword_type swlength = ( ( dsp_image.channels() * dsp_image.cols ) * dsp_image.rows );
    m_out_img.images[seq_num].width  = dsp_image.cols;
    m_out_img.images[seq_num].height = dsp_image.rows;
    m_out_img.images[seq_num].bpp    = FORMAT_COLOR24;
    m_out_img.images[seq_num].format = "bitmap";
    m_out_img.images[seq_num].fDiv   = 0;
    m_out_img.images[seq_num].pixels.length( swlength );
    ::memcpy( ( void * )&( m_out_img.images[ seq_num ].pixels[ 0 ] ) , dsp_image.data , swlength );

    auwReturn = RESULT_OK;
  }

  return auwReturn;
}

/***************************************************************************************************
 *  Function        : bAllocateMemory_m_corsp
 *  Description     : This function allocates the area of the m_corsp variable.
 *  Argument        : reference image length
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_factorization::bAllocateMemory_m_corsp( uword_type ref_len )
{
  bool_type abReturn = FALSE;
  bool_type bm_corsp = TRUE;

  try
  {
    rtc_m_corsp = new rtcPos2D[ ref_len ];
  }
  catch( std::bad_alloc )
  {
    std::cout << _ERROR_HEAP_MEM_MCORSP << std::endl;
    bm_corsp = FALSE;
  }

  if( bm_corsp )
  {
    memset( rtc_m_corsp, 0, sizeof( rtcPos2D ) * ref_len );
    abReturn = TRUE;
  }

  return abReturn;

}

/***************************************************************************************************
 *  Function        : bAllocateMemory_pickbuf
 *  Description     : This function allocates the area of the pickbuf variable.
 *  Argument        : input image length
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_factorization::bAllocateMemory_pickbuf( uword_type inp_len )
{
  bool_type abReturn = FALSE;
  bool_type bpickbuf = TRUE;

  try
  {
    pickbuf = new char_type[ inp_len ];
  }
  catch( std::bad_alloc )
  {
    std::cout << _ERROR_HEAP_MEM_PICKBUF << std::endl;
    bpickbuf = FALSE;
  }

  if( bpickbuf )
  {
    m_rlt_2dcdp.pickbuf.length(inp_len);
    memset( pickbuf, 0, sizeof( char_type ) * inp_len );
    m_rlt_2dcdp.pickbuf.release();
    abReturn = TRUE;
  }

  return abReturn;

}
/***************************************************************************************************
 *  Function        : bAllocateMemory_pickup1
 *  Description     : This function allocates the area of the pickup1 variable.
 *  Argument        : reference image length
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_factorization::bAllocateMemory_pickup1( uword_type ref_len )
{
  bool_type abReturn = TRUE;

  try
  {
    pickup1 = new rtcPICK[ ref_len ];
  }
  catch( std::bad_alloc )
  {
    std::cout << _ERROR_HEAP_MEM_PICKUP1 << std::endl;
    abReturn = FALSE;
  }

  return abReturn;

}

/***************************************************************************************************
 *  Function        : bAllocateMemory_pickup2
 *  Description     : This function allocates the area of the pickup2 variable.
 *  Argument        : reference image length
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_factorization::bAllocateMemory_pickup2( uword_type ref_len )
{
  bool_type abReturn = TRUE;

  try
  {
    pickup2 = new rtcPICK[ ref_len ];
  }
  catch( std::bad_alloc )
  {
    std::cout << _ERROR_HEAP_MEM_PICKUP2 << std::endl;
    abReturn = FALSE;
  }

  return abReturn;

}

/***************************************************************************************************
 *  Function        : nvdCopy_pickup1
 *  Description     : This function copies data from "m_rlt_2dcdp.pickup" to "pickup1".
 *  Argument        : reference image length
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdCopy_pickup1( uword_type ref_len )
{
  m_rlt_2dcdp.pickup.length(ref_len);

  uword_type i = 0;
  for (i = 0; i < ref_len; i++)
  {
    pickup1[i].x  = m_rlt_2dcdp.pickup[i].x;
    pickup1[i].y  = m_rlt_2dcdp.pickup[i].y;
    pickup1[i].x0 = m_rlt_2dcdp.pickup[i].x0;
    pickup1[i].y0 = m_rlt_2dcdp.pickup[i].y0;
  }

  m_rlt_2dcdp.pickup.release();

}

/***************************************************************************************************
 *  Function        : nvdCopy_pickup2
 *  Description     : This function copies data from "m_rlt_2dcdp.pickup" to "pickup2".
 *  Argument        : reference image length
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdCopy_pickup2( uword_type ref_len )
{
  m_rlt_2dcdp.pickup.length(ref_len);

  uword_type i = 0;
  for (i = 0; i < ref_len; i++)
  {
    pickup2[i].x  = m_rlt_2dcdp.pickup[i].x;
    pickup2[i].y  = m_rlt_2dcdp.pickup[i].y;
    pickup2[i].x0 = m_rlt_2dcdp.pickup[i].x0;
    pickup2[i].y0 = m_rlt_2dcdp.pickup[i].y0;
  }

  m_rlt_2dcdp.pickup.release();

}

/***************************************************************************************************
 *  Function        : nvdSetStageControl
 *  Description     : This function sets the member variable of m_ubStageCtrl.
 *  Argument        : m_ubStageCtrl
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdSetStageControl( ubyte_type a_ubStageCtrl )
{
  if( (a_ubStageCtrl >= FACTOR_STAGE_MIN) && (a_ubStageCtrl < FACTOR_STAGE_MAX) )
  {
    m_ubStageCtrl = a_ubStageCtrl;
  }
  else
  {
    m_ubStageCtrl = FACTOR_STAGE1;
  }
}

/***************************************************************************************************
 *  Function        : nvdSetErrorCode
 *  Description     : This function sets the member variable of m_uhwErrorCode.
 *  Argument        : m_uhwErrorCode
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdSetErrorCode( uhword_type a_uhwErrorCode )
{
  if( (a_uhwErrorCode >= ERR_CODE_MIN) && (a_uhwErrorCode < ERR_CODE_MAX) )
  {
    m_uhwErrorCode = a_uhwErrorCode;
  }
  else
  {
    m_uhwErrorCode = ERR_CODE_INIT;
  }
}

/***************************************************************************************************
 *  Function        : nvdOutputErrorMessages
 *  Description     : This function is error messages output to the console.
 *  Argument        : m_uhwErrorCode
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdOutputErrorMessages( void )
{
  switch ( m_uhwErrorCode )
  {
    case ERR_IMAGE_DATA:
      std::cout << _ERROR_IMAGE_DATA << std::endl;
      std::cout << _ERROR_RESET << std::endl;
      break;
    case ERR_ALLOCATE_M:
      std::cout << _ERROR_ALLOCATION << std::endl;
      std::cout << _ERROR_RESET << std::endl;
      break;
    case ERR_OBJ_CREATE:
      std::cout << _ERROR_OBJ_CREATE << std::endl;
      std::cout << _ERROR_RESET << std::endl;
      break;

    default:
      break;
  }
  nvdSetErrorCode( ERR_CODE_INIT );

}

/***************************************************************************************************
 *  Function        : bCreateMesh
 *  Description     : This function execution a factorization method process.
 *  Argument        : object data
 *  Return          : object data
 **************************************************************************************************/
bool_type rtc_factorization::bCreateMesh( OBJ_MESH::Mesh &mesh )
{
  bool_type abFlag = FALSE;

  TrajectData             trajectdata( NUM_IMAGE );
  FactorizationController fcontroller;
  CDPObjFormatter         cdpformatter;
  bgColor = CDP::Color( 0, 0, 255 );

  uword_type pixnum = 0;
  IplImage* temp = cvCreateImage( cvSize(XIMG_120, YIMG_90), IPL_DEPTH_8U, CHANNEL_NUM );

  #ifdef _DEBUG_SW_
  LARGE_INTEGER	freq, time_start, time_end;
  QueryPerformanceFrequency( &freq );
  QueryPerformanceCounter( &time_start );
  #endif

  if( bSetTrajectData1( trajectdata, LEFT ) )
  {
    if( bSetTrajectDataCenter( trajectdata, CENTER ) )
    {
      mesh.SetTextureImage( ref.iplImage() );
      if( bSetTrajectData2( trajectdata, RIGHT ) )
      {
        pixnum = m_pix_num;
        abFlag = TRUE;
      }
    }
  }

  if( abFlag == FALSE ) return FALSE;

/***************************************************************************************************
 **************************************************************************************************/

  fcontroller.set( trajectdata.GetTrajectVector() );

  for( uword_type i = 0; i < NUM_IMAGE; i++)
  {
    fcontroller.createData( trajectdata, i );
  }
  
  /* Save the trajectory data and output the filename to data port. */
  nvdTrajectFileCreation(trajectdata, pixnum);
 
  /* Output the trajectory data to data port. */
  nvdResultTrajectCreation(trajectdata, pixnum);

  cdpformatter.setImageSize( temp, temp );
  cdpformatter.create( trajectdata, CENTER, pixnum );

  mesh.Init();

  if( !cdpformatter.writeForFile( mesh ) )
  {
    std::cout << _ERROR_OBJECT_CREATION << std::endl;
  }

  #ifdef _DEBUG_SW_
  QueryPerformanceCounter( &time_end );
  std::cout << "Process Time of Factorization-> " << ((time_end.QuadPart - time_start.QuadPart)*1000 / freq.QuadPart) << "[ms]" << std::endl;
  #endif

  //#ifdef _DEBUG_SW_
  std::stringstream filename;
  filename << "factorization" << std::setw(2) << std::setfill('0') << m_uw_fileno << ".obj" << std::flush;
  mesh.SaveObjectFile(filename.str().c_str());
  //#endif

  nvdResultObjectCreation( mesh );

  /* Initialize dynamic memory */
  trajectdata.Finish();
  cdpformatter.release();
  nvdRelease();
  cvReleaseImage( &temp );

  return TRUE;
}

/***************************************************************************************************
 *  Function        : nvdResultObjectCreation
 *  Description     : This function sets object data to be output port.
 *  Argument        : object data
 *  Return          : output port
 **************************************************************************************************/
void rtc_factorization::nvdResultObjectCreation( OBJ_MESH::Mesh &mesh )
{
  nvdResultObjectVertex( mesh );
  nvdResultObjectTexture( mesh );
  nvdResultObjectFace( mesh );

  bool_type abWrtResult = m_rlt_objOut.write();	
  if( abWrtResult == FALSE )
  {
    std::cout << _ERROR_WRITE_OBJ << std::endl;
  }

  /* Initialize dynamic memory */
  m_rlt_obj.p_vertex.release();
  m_rlt_obj.p_uv.release();
  m_rlt_obj.p_face.release();

}

/***************************************************************************************************
 *  Function        : nvdResultObjectVertex
 *  Description     : This function sets the vertex data to the output port.
 *  Argument        : vertex data
 *  Return          : output port
 **************************************************************************************************/
void rtc_factorization::nvdResultObjectVertex( OBJ_MESH::Mesh &mesh )
{
  uword_type a_vertex_idx_len = mesh.GetVertexNum();
  m_rlt_obj.vertex_idx_len = a_vertex_idx_len;
  m_rlt_obj.p_vertex.length(a_vertex_idx_len);

  m_rlt_obj.vertex_num_len = mesh.GetVertex_Num();

  uword_type i = 0;
  for (i = 0; i < a_vertex_idx_len; i++)
  {
    m_rlt_obj.p_vertex[i].x = mesh.GetVertex(i).x;
    m_rlt_obj.p_vertex[i].y = mesh.GetVertex(i).y;
    m_rlt_obj.p_vertex[i].z = mesh.GetVertex(i).z;
  }
}

/***************************************************************************************************
 *  Function        : nvdResultObjectTexture
 *  Description     : This function sets the texture data(UV) to the output port.
 *  Argument        : texture data
 *  Return          : output port
 **************************************************************************************************/
void rtc_factorization::nvdResultObjectTexture( OBJ_MESH::Mesh &mesh )
{
  uword_type a_uv_idx_len = mesh.GetUVNum();
  m_rlt_obj.uv_idx_len = a_uv_idx_len;
  m_rlt_obj.p_uv.length(a_uv_idx_len);

  m_rlt_obj.uv_num_len = mesh.GetUV_Num();

  uword_type i = 0;
  for (i = 0; i < a_uv_idx_len; i++)
  {
    m_rlt_obj.p_uv[i].u = mesh.GetUV(i).u;
    m_rlt_obj.p_uv[i].v = mesh.GetUV(i).v;
  }
}

/***************************************************************************************************
 *  Function        : nvdResultObjectFace
 *  Description     : This function sets the face data to the output port.
 *  Argument        : face data
 *  Return          : output port
 **************************************************************************************************/
void rtc_factorization::nvdResultObjectFace( OBJ_MESH::Mesh &mesh )
{
  uword_type a_face_idx_len = mesh.GetFaceNum();
  m_rlt_obj.face_idx_len = a_face_idx_len;
  m_rlt_obj.p_face.length(a_face_idx_len);

  m_rlt_obj.face_size = mesh.GetFace_Size();
  m_rlt_obj.depth_max = mesh.GetMaxDepth();
  m_rlt_obj.depth_min = mesh.GetMinDepth();

  m_rlt_obj.face_num_len = mesh.GetFace_Num();

  uword_type i = 0;
  for( i = 0; i < a_face_idx_len; i++ )
  {
    m_rlt_obj.p_face[i].element = mesh.GetFaceElement(i);
    m_rlt_obj.p_face[i].type = (static_cast<unsigned long>(mesh.GetFaceType(i)));

    uword_type j = 0;
    for( j = 0; j < m_rlt_obj.p_face[i].element; j++ )
    {
      m_rlt_obj.p_face[i].vertex_idx[j] = mesh.GetFaceVertex_index(i,j);
    }
  }
}

/***************************************************************************************************
 *  Function        : bSetTrajectData1
 *  Description     : This function sets the result of 2dcdp1 to traject-data.
 *  Argument        : result of 2dcdp1
 *  Return          : TrajectData
 **************************************************************************************************/
bool_type rtc_factorization::bSetTrajectData1( TrajectData& data, const int& index )
{
  udword_type dm = refW;
  udword_type di = inpW;

  if( rtc_m_corsp == NULL ){
    return FALSE;
  }

  memset( rtc_m_corsp, 0, sizeof( rtcPos2D ) * refW * refH );
  m_pix_num = 0;

  for( udword_type n = 0; n < refH; n++ ){
    for( udword_type m = 0; m < refW; m++ ){
      if( ref[(uword_type)n][(uword_type)m] != bgColor ){
        pickbuf[pickup1[n*dm+m].y*di + pickup1[n*dm+m].x] = 1;

        rtc_m_corsp[refW * n + m].i = pickup1[n*dm + m].x;
        rtc_m_corsp[refW * n + m].j = pickup1[n*dm + m].y;
        rtc_m_corsp[refW * n + m].isData = true;

        m_pix_num++;
      }
    }
  }

  if( !data.Malloc( m_pix_num ) ){
    return FALSE;
  }

  uword_type n = 0;
  for( udword_type y = 0; y < refH; y++ ){
    for( udword_type x = 0; x < refW; x++ ){
      if( rtc_m_corsp[refW * y + x].isData ){
        data.SetTrajectData( 2*index, n, rtc_m_corsp[refW * y + x].i );
        data.SetTrajectData( 2*index+1, n, rtc_m_corsp[refW * y + x].j );

        data.SetIJMNXYZ( index, n ).i = rtc_m_corsp[refW * y + x].i;
        data.SetIJMNXYZ( index, n ).j = rtc_m_corsp[refW * y + x].j;
        data.SetIJMNXYZ( index, n ).m = (int)x;
        data.SetIJMNXYZ( index, n ).n = (int)y;
        n++;
      }
    }
  }

  return TRUE;

}

/***************************************************************************************************
 *  Function        : bSetTrajectData2
 *  Description     : This function sets the result of 2dcdp2 to traject-data.
 *  Argument        : result of 2dcdp2
 *  Return          : TrajectData
 **************************************************************************************************/
bool_type rtc_factorization::bSetTrajectData2( TrajectData& data, const int& index )
{
  udword_type dm = refW;
  udword_type di = inpW;

  if( rtc_m_corsp == NULL ){
    return FALSE;
  }

  memset( rtc_m_corsp, 0, sizeof( rtcPos2D ) * refW * refH );
  m_pix_num = 0;

  for( udword_type n = 0; n < refH; n++ ){
    for( udword_type m = 0; m < refW; m++ ){
      if( ref[(uword_type)n][(uword_type)m] != bgColor ){
        pickbuf[pickup2[n*dm+m].y*di + pickup2[n*dm+m].x] = 1;

        rtc_m_corsp[refW * n + m].i = pickup2[n*dm + m].x;
        rtc_m_corsp[refW * n + m].j = pickup2[n*dm + m].y;
        rtc_m_corsp[refW * n + m].isData = true;

        m_pix_num++;
      }
    }
  }

  if( !data.Malloc( m_pix_num ) ){
    return FALSE;
  }

  uword_type n = 0;
  for( udword_type y = 0; y < refH; y++ ){
    for( udword_type x = 0; x < refW; x++ ){
      if( rtc_m_corsp[refW * y + x].isData ){
        data.SetTrajectData( 2*index, n, rtc_m_corsp[refW * y + x].i );
        data.SetTrajectData( 2*index+1, n, rtc_m_corsp[refW * y + x].j );

        data.SetIJMNXYZ( index, n ).i = rtc_m_corsp[refW * y + x].i;
        data.SetIJMNXYZ( index, n ).j = rtc_m_corsp[refW * y + x].j;
        data.SetIJMNXYZ( index, n ).m = (int)x;
        data.SetIJMNXYZ( index, n ).n = (int)y;
        n++;
      }
    }
  }

  return TRUE;

}

/***************************************************************************************************
 *  Function        : bSetTrajectDataCenter
 *  Description     : 
 *  Argument        : TrajectData
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_factorization::bSetTrajectDataCenter(TrajectData& data, const int& index)
{
  uword_type n = 0;
  m_pix_num    = 0;

  for( udword_type y = 0; y < refH; y++ )
    for( udword_type x = 0; x < refW; x++ )
      if( ref[(uword_type)y][(uword_type)x] != bgColor )  m_pix_num++;

  if( !data.Malloc( m_pix_num ) ){
    return FALSE;
  }

  for( udword_type y = 0; y < refH; y++ ){
    for( uword_type x = 0; x < refW; x++ ){
      if( ref[(uword_type)y][(uword_type)x] != bgColor ){
        data.SetTrajectData( 2*index, n, (int)x );
        data.SetTrajectData( 2*index+1, n, (int)y );

        data.SetIJMNXYZ( index, n ).i = (int)x;
        data.SetIJMNXYZ( index, n ).j = (int)y;
        data.SetIJMNXYZ( index, n ).m = (int)x;
        data.SetIJMNXYZ( index, n ).n = (int)y;
        n++;
      }
    }
  }

  return TRUE;

}

/***************************************************************************************************
 *  Function        : nvdRelease
 *  Description     : This function releases dynamic memory.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_factorization::nvdRelease( void )
{
  if( rtc_m_corsp != NULL ){
    delete [] rtc_m_corsp;
    rtc_m_corsp = NULL;
  }

  if( pickbuf != NULL ){
    delete [] pickbuf;
    pickbuf = NULL;
  }

  if( pickup1 != NULL ){
    delete [] pickup1;
    pickup1 = NULL;
  }

  if( pickup2 != NULL ){
    delete [] pickup2;
    pickup2 = NULL;
  }
}

/***************************************************************************************************
*  Function        : nvdTrajectFileCreation
*  Description     : This function saves the trajectory file and outputs the file name to the data port.
*  Argument        : TrajectData
*  Return          : nothing special
**************************************************************************************************/
void rtc_factorization::nvdTrajectFileCreation(const TrajectData& data, uword_type uw_num)
{
  std::stringstream filename[NUM_IMAGE];
  ubyte_type ub_position[NUM_IMAGE]{'L', 'C', 'R'};
  m_uw_fileno = DEF_FILE_NO;

  if (m_trj_filenoIn.isNew())
  {
    m_trj_filenoIn.read();
    m_uw_fileno = m_trj_fileno.data;
  }
 
  for (uword_type i = 0; i < NUM_IMAGE; i++)
  {
    /* Create a file name. */
    filename[i] << g_absolute_dir << "\TrajectData" << std::setw(2) << std::setfill('0') << m_uw_fileno << ub_position[i] << ".ijmnxyz" << std::flush;

    if (!bSaveTraject(data, filename[i].str().c_str(), i, uw_num))
    {
      std::cout << _ERROR_TRJECT_CREATION << std::endl;
      m_out_trjpath.data[i] = _ERROR_FILE_CREATION;
    }
    else
    {
      m_out_trjpath.data[i] = filename[i].str().c_str();
    }
  }
  
  bool_type abWrtResult = m_out_trjpathOut.write();
  if (abWrtResult == FALSE)
  {
    std::cout << _ERROR_WRITE_TRJPATH << std::endl;
  }
}

/***************************************************************************************************
*  Function        : bSaveIJMNXYZ
*  Description     : This function saves the trajectory file.
*  Argument        : TrajectData
*  Return          : nothing special
**************************************************************************************************/
bool_type rtc_factorization::bSaveTraject(const TrajectData& data, const char *filename, uword_type uw_index, uword_type uw_num)
{
  TrajectData::IJMNXYZ temp;
  FILE* fp;

  if (filename == NULL){
    return FALSE;
  }

  fp = fopen(filename, "wb");
  if (fp == NULL){
    return FALSE;
  }

  fprintf(fp, "# Created by IJMNXYZ data\n");
  fprintf(fp, "# Length : %d\n", uw_num);
  fprintf(fp, "# Data :\n");

  for (uword_type count = 0; count < uw_num; count++)
  {
    temp = data.GetIJMNXYZ(uw_index, count);
    fprintf(fp, "%d %d %d %d %.6f %.6f %.6f\n", temp.i, temp.j, temp.m, temp.n, temp.x, temp.y, temp.z);
  }

  fclose(fp);

  return TRUE;
}

/***************************************************************************************************
*  Function        : nvdResultTrajectCreation
*  Description     : This function sets trajectory data to be output port.
*  Argument        : object data
*  Return          : output port
**************************************************************************************************/
void rtc_factorization::nvdResultTrajectCreation(const TrajectData& data, uword_type uw_num)
{
  for (uword_type i = 0; i < NUM_IMAGE; i++)
  {
    nvdSetTraject(data, i, uw_num);
  }

  bool_type abWrtResult = m_out_trjOut.write();
  if (abWrtResult == FALSE)
  {
    std::cout << _ERROR_WRITE_TRJ << std::endl;
  }
  
  for (uword_type i = 0; i < NUM_IMAGE; i++)
  {
    m_out_trj.data[i].release();
  }

}

/***************************************************************************************************
*  Function        : nvdSetTraject
*  Description     : This function sets the trajectory data.                   
*  Argument        : TrajectData
*  Return          : nothing special
**************************************************************************************************/
void rtc_factorization::nvdSetTraject(const TrajectData& data, uword_type uw_index, uword_type uw_num)
{
  TrajectData::IJMNXYZ temp;

  m_out_trj.data[uw_index].length(uw_num);
  m_out_trj.data_len = uw_num;

  for (uword_type count = 0; count < uw_num; count++)
  {
    temp = data.GetIJMNXYZ(uw_index, count);
    m_out_trj.data[uw_index][count].i = temp.i;
    m_out_trj.data[uw_index][count].j = temp.j;
    m_out_trj.data[uw_index][count].m = temp.m;
    m_out_trj.data[uw_index][count].n = temp.n;
    m_out_trj.data[uw_index][count].x = temp.x;
    m_out_trj.data[uw_index][count].y = temp.y;
    m_out_trj.data[uw_index][count].z = temp.z;
  }
}

/***************************************************************************************************
 **************************************************************************************************/

extern "C"
{
 
  void rtc_factorizationInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_factorization_spec);
    manager->registerFactory(profile,
                             RTC::Create<rtc_factorization>,
                             RTC::Delete<rtc_factorization>);
  }
  
};


